#include "adc.h"

/* ############################################################################# */

#include "touch.h"

#define ADC_BUFFER_LEN	4

#define PRESS_STATE_0	0x00
#define PRESS_STATE_1	0x01
#define PRESS_STATE_2	0x02
#define PRESS_STATE_3	0x03
#define X_STATE_0	0x10
#define X_STATE_1	0x11
#define X_STATE_2	0x12
#define Y_STATE_0	0x20
#define Y_STATE_1	0x21
#define Y_STATE_2	0x22

static struct coord {
	int16_t pressed;	// TOUCH_PRESSED_LEVEL wenn losgelassen
	int16_t released;	// TOUCH_PRESSED_LEVEL wenn losgelassen
	int16_t x;		// nur gueltig wenn pressed UND relased < TOUCH_PRESSED_LEVEL
	int16_t y;
} touch, ret;

#define ADPS_DIV2	(_BV(ADPS0))
#define ADPS_DIV4	(_BV(ADPS1))
#define ADPS_DIV8	(_BV(ADPS1)|_BV(ADPS0))
#define ADPS_DIV16	(_BV(ADPS2))
#define ADPS_DIV32	(_BV(ADPS2)|_BV(ADPS0))
#define ADPS_DIV64	(_BV(ADPS2)|_BV(ADPS1))
#define ADPS_DIV128	(_BV(ADPS2)|_BV(ADPS1)|_BV(ADPS0))

/* ADC sample frequency should be between 50-200 KHz.  Pick the fastest
 * sample frequency (smallest divisor) we can use ... we know that the
 * lowest clock is 200 KHz (for I2C), so we can never fail here.
 */
static inline uint8_t adc_ps(void)
{
	#define div (F_CPU/200000)

	if (div <= 2)
		return ADPS_DIV2;
	else if (div > 2 && div <= 4)
		return ADPS_DIV4;
	else if (div > 4 && div <= 8)
		return ADPS_DIV8;
	else if (div > 8 && div <= 16)
		return ADPS_DIV16;
	else if (div > 16 && div <= 32)
		return ADPS_DIV32;
	else if (div > 32 && div <= 64)
		return ADPS_DIV64;
	else
		return ADPS_DIV128;
}

#if F_CPU <= 1000000
#define TIMER_0_CLOCK (F_CPU/250)
#define T0_PRESCALER 	1	// F_CPU/4000/1 <= 250; passt in ein Byte
#elif F_CPU <= 8000000
#define TIMER_0_CLOCK (F_CPU/250/8)
#define T0_PRESCALER 	8	// F_CPU/4000/8 <= 250; passt in ein Byte
#elif F_CPU <= 16000000
#define TIMER_0_CLOCK (F_CPU/250/64)
#define T0_PRESCALER 	64	// F_CPU/4000/64 <= 250; passt in ein Byte
#endif

static uint8_t adc_state   = 0;	// 
				
// Initialisert den AD-Umsetzer. Fr jeden Kanal, den man nutzen mchte, 
// muss das entsprechende Bit in channel gesetzt sein:
// Bit0 = Kanal 0 usw.
void adc_init(void)
{

	TCCR0 = 0;	// stop counter // Normal operation Output pins disconnected
	TCNT0 = 0;	//reset counter

	// TIMER_x_CLOCK muss kleiner/gleich als 4kHz sein (Siehe unten: ADC Timing mit max 250uS conversion time)
	// set timer TOP		
	OCR0 = ((F_CPU/T0_PRESCALER/TIMER_0_CLOCK) - 1 );	
	
	// clear OC0 interrupt flag
	TIFR =  _BV(OCF0);			// Reset OCF0 - Flag
	// ndert man den Prescaler muss man die Formel fr OCR0 anpassen !!!
	// use CLK/8 prescale value, clear timer/counter on compare match   
#if T0_PRESCALER == 1
	TCCR0 = _BV(CS00) | _BV(WGM01);
#elif T0_PRESCALER == 8
	TCCR0 = _BV(CS01) | _BV(WGM01);
#elif T0_PRESCALER == 64
	TCCR0 = _BV(CS01) | _BV(CS00)  | _BV(WGM01);
#elif T0_PRESCALER == 256
	TCCR0 = _BV(CS02) | _BV(WGM01);
#elif T0_PRESCALER == 1024
	TCCR0 = _BV(CS02) | _BV(CS00)  | _BV(WGM01);
#else
	#error Invalid T0_PRESCALER value. Valid is 1, 8, 64, 256 or 1024
#endif	

	touch.pressed = touch.released = TOUCH_PRESSED_LEVEL;
	
	// AVcc als Versorgungspannung, rechts Ausrichtung
	ADMUX |= (1 << REFS0);

	SFIOR |= (1 << ADTS1) | (1 << ADTS0);	// ADC Trigger Source = Timer0 Compare Match
	
	TOUCH_DDR |=  (1 << TOUCH_X1);
	TOUCH_DDR |=  (1 << TOUCH_Y1);
	TOUCH_DDR &= ~(1 << TOUCH_X2);

	TOUCH_PORT &= ~(1 << TOUCH_Y1);
	TOUCH_PORT |=  (1 << TOUCH_X1);	
	
	adc_state = PRESS_STATE_0;		// Zeiger in den Puffer
	
	// Wir fangen an mit Press_state 
	ADMUX = (ADMUX & 0xE0) | TOUCH_X2;	// Und jetzt Kanal whlen

	ADCSRA = adc_ps() |				// Pick the fastest sample frequency (smallest divisor) we can use 
		 _BV(ADEN)|			// ADC an
		 _BV(ADSC)|			// Beginne mit der Konvertierung
		 _BV(ADIE)|			// ADC-Interrupt an
		 _BV(ADATE);			// Auto trigger on
	TIFR =  _BV(OCF0);			// Reset OCF0 - Flag
}

/* Interrupt, ausgelst, wenn der AD-Wandler fertig ist 
 * speichert einen Messwert. (vorher wendet sie evtl. noch eine Filterfkt an)
 */
ISR(ADC_vect)
{
	int16_t adc = ADCW;			// ADC einlesen

	TIFR =  _BV(OCF0);			// Reset OCF0 - Flag
	switch(adc_state)
	{
	case PRESS_STATE_0:			// Erste Wandlung "PRESS_STATE" nach Kanalumschaltung
		adc_state = PRESS_STATE_1;
		//ADC war Dummy
		break;
	case PRESS_STATE_1:			// Gueltige "PRESS_STATE" Wandlung
		if(adc >= TOUCH_PRESSED_LEVEL)
		{
			touch.pressed = touch.released = adc;
			ret = touch;
		}
		touch.pressed = adc;
		adc_state = X_STATE_0;
		TOUCH_DDR |=  (1 << TOUCH_X1);
		TOUCH_DDR |=  (1 << TOUCH_X2);
		TOUCH_DDR &= ~(1 << TOUCH_Y1);
		TOUCH_DDR &= ~(1 << TOUCH_Y2);

		TOUCH_PORT &= ~(1 << TOUCH_X1);
		TOUCH_PORT |=  (1 << TOUCH_X2);
		ADMUX = (ADMUX & 0xE0) | TOUCH_Y1;	// ADC Kanal fr X-Messung
		break;
	case X_STATE_0:		// X dummy read
		adc_state = X_STATE_1;
		break;
	case X_STATE_1:		// erster gueltiger X-Messwert
		adc_state = X_STATE_2;
		touch.x = adc;
		break;
	case X_STATE_2:		// Zweiter gueltiger X-Messwert
		if(((touch.x > adc) && ((touch.x - adc) > 10))
		|| ((adc - touch.x) > 10))
		{	// nochmal. Letzen Wert als Vergleichswert merken
			touch.x = adc;		// Erster X und zweiter X liegen zuweit auseinander
		}
		else	// Die zwei X Messwerte liegen nahe beieinander 
		{
			touch.x += adc;
			adc_state = Y_STATE_0;
			TOUCH_DDR |=  (1 << TOUCH_Y1);
			TOUCH_DDR |=  (1 << TOUCH_Y2);
			TOUCH_DDR &= ~(1 << TOUCH_X1);
			TOUCH_DDR &= ~(1 << TOUCH_X2);

			TOUCH_PORT |=  (1 << TOUCH_Y1);
			TOUCH_PORT &= ~(1 << TOUCH_Y2);

			ADMUX = (ADMUX & 0xE0) | TOUCH_X1;	// ADC Kanal fr Y-Messung
		}

		break;
	case Y_STATE_0:		// Y dummy read
		adc_state = Y_STATE_1;
		break;
	case Y_STATE_1:
		adc_state = Y_STATE_2;
		touch.y = adc;
		break;
	case Y_STATE_2:
		if(((touch.y > adc) && ((touch.y - adc) > 10))	// zu viel jitter
		|| ((adc - touch.y) > 10))
		{
			touch.y = adc;	// nochmal. Letzen Wert als Vergleichswert merken
		}
		else		// OK gueltiger Wert
		{
			touch.y += adc;
			TOUCH_DDR |=  (1 << TOUCH_X1);
			TOUCH_DDR |=  (1 << TOUCH_Y1);
			TOUCH_DDR &= ~(1 << TOUCH_X2);

			TOUCH_PORT &= ~(1 << TOUCH_Y1);
			TOUCH_PORT |=  (1 << TOUCH_X1);	
			adc_state = PRESS_STATE_2;
			ADMUX = (ADMUX & 0xE0) | TOUCH_X2;	// ADC Kanal fr Pressed-Messung
		}

		break;
	case PRESS_STATE_2:		// Am Ende der Messungen prfen ob Messung gltig
		adc_state = PRESS_STATE_3;
		break;
	case PRESS_STATE_3:		// Am Ende der Messungen prfen ob Messung gltig
		if(adc >= TOUCH_PRESSED_LEVEL)
		{
			touch.pressed = touch.released = adc;
			ret = touch;
			break;		// released
		}
		touch.released = adc;

		// Prepare return value
		ret = touch;

		touch.pressed = touch.released = TOUCH_PRESSED_LEVEL+1;

		adc_state = PRESS_STATE_1;
		break;
	}
}

uint16_t readADC(uint8_t channel) 
{
	int16_t retval = -1;

	ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
	{
	
	switch (channel)
	{
	case TOUCH_X2:	// Press
		retval = ((ret.pressed < TOUCH_PRESSED_LEVEL) && (ret.released < TOUCH_PRESSED_LEVEL)) ? ret.pressed : TOUCH_PRESSED_LEVEL+1;
		break;
	case TOUCH_Y1:	// X
		retval = ((ret.pressed < TOUCH_PRESSED_LEVEL) && (ret.released < TOUCH_PRESSED_LEVEL)) ? ret.x : -1;
		break;
	case TOUCH_X1:	// Y
		retval = ((ret.pressed < TOUCH_PRESSED_LEVEL) && (ret.released < TOUCH_PRESSED_LEVEL)) ? ret.y : -1;
		break;
	}
	}
	return retval;
}

