/*
 * lc7981.c
 *
 *  Created on: 01.05.2009
 *      Author: sebastian
 *
 *    This file is part of Sebastians AVR Library for lc7981.
 *
 *   Sebastians AVR Library is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Sebastians AVR Library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Sebastians AVR Library.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "lc7981.h"

#define nop() asm volatile ("nop" ::)

static inline void delay_140ns(void)
{
//	nop();
#if (F_CPU > 7142857UL)
	nop();
#endif
#if (F_CPU > 2*7142857UL)
	nop();
#endif
#if (F_CPU > 3*7142857UL)
	nop();
#endif
}


static inline void delay_90ns(void)
{
//	nop();
#if (F_CPU > 11111111UL)
	nop();
#endif
#if (F_CPU > 2*11111111UL)
	nop();
#endif
}

static inline void delay_220ns(void)
{
//	nop();
#if (F_CPU > 4545454UL)
	nop();
#endif
#if (F_CPU > 2*4545454UL)
	nop();
#endif
#if (F_CPU > 3*4545454UL)
	nop();
#endif
#if (F_CPU > 4*4545454UL)
	nop();
#endif
}


uint8_t lcd_mode, 		//!< Currently used mode, can be text or graphic.
		lcd_curline;	//!< Current line. Used for scrolling in text mode.

static inline void lcd_waitbusy(void)
{
	unsigned char temp;
	
	LCD_DIR_READ();
	lcd_rw_read();
	lcd_rs_instruction();
	do
	{
		delay_90ns();	// Address setup time
		lcd_en_active();
		delay_140ns();	// Data delay time (read) max 140ns
		temp = LCD_DATA_PIN;
		lcd_en_passive();
	}
	while (temp & (1<<BUSYBIT));

	lcd_rw_write();
	LCD_DIR_WRITE();
}


/**
 * Generates the strobe signal for writing data.
 * This function is meant for internal usage only.
 */
static inline void lcd_strobe(uint8_t cmd) 
{
	delay_90ns();		// Address setup time
	LCD_DATA = cmd;
	lcd_en_active();
	delay_220ns();		// Data setup time
	lcd_en_passive();
}


/**
 *  Writes a command and a data byte to the lcd.
 *
 *  @param cmd the command byte
 *  @param data the data that is going to be written after the command
 */
static inline void lcd_write_command(uint8_t cmd, uint8_t data) 
{
	lcd_waitbusy();
	lcd_strobe(cmd);
	lcd_rs_data();
	lcd_strobe(data);
}


/**
 * lcd_init set the to default values.
 *
 * @param mode LCD_TEXT_MODE for text mode and LCD_GRAPHIC_MODE for graphic mode
 */
void lcd_init(uint8_t mode) 
{
	LCD_DATA     = 0;
	LCD_CTRL     = 0;
	LCD_CRTL_DDR |= (1 << LCD_RS) | (1 << LCD_RW) | (1 << LCD_EN) | (1<<LCD_CS) | (1<<LCD_RESET);
	_delay_us(100);
	LCD_CTRL     |= (1<<LCD_RESET);
	_delay_ms(100);


	lcd_en_passive();
	lcd_rs_data();
	lcd_rw_write();
	LCD_DIR_WRITE();

	if(mode == LCD_TEXT_MODE) 
	{
		lcd_mode = LCD_TEXT_MODE;

		lcd_write_command(0x00,0x3C);
		lcd_write_command(0x01,0x75);
		lcd_write_command(0x02,LCD_TEXT_COLUMNS - 1);
		lcd_write_command(0x03,0x4F);
		lcd_write_command(0x04,0x07);
		lcd_write_command(0x08,0x00);
		lcd_write_command(0x09,0x00);
		lcd_write_command(0x0A,0x00);
		lcd_write_command(0x0B,0x00);

	}
	else 
	{
		lcd_mode = LCD_GRAPHIC_MODE;

		lcd_write_command(0x00,0x32);
		lcd_write_command(0x01,0x07);
		lcd_write_command(0x02,(LCD_GRAPHIC_WIDTH / 8)-1);
		lcd_write_command(0x03,80-1);
		lcd_write_command(0x08,0x00);
		lcd_write_command(0x09,0x00);
		lcd_write_command(0x0A,0x00);
		lcd_write_command(0x0B,0x00);
	}

	lcd_clear();
	BACKLIGHT	|= (1<<BACKLIGHT_EN);
	BACKLIGHT_DDR	|= (1<<BACKLIGHT_EN);
}

/**
 * Reads a byte from the display memory.
 * lcd_gotoxy can be used to set the location.
 * Important : lcd_gotoxy doesn't work in graphics mode yet.
 *
 * @return the byte which has been read
 * @see lcd_gotoxy
 */
inline uint8_t lcd_read_byte(void) 
{
	uint8_t i,data;

	for(i = 0; i < 2; i++) 
	{
		lcd_waitbusy();
		LCD_DATA = 0x0D;
		delay_90ns();	// Address setup time
		lcd_en_active();
		lcd_rs_data();
		lcd_rw_read();
		LCD_DIR_READ();
		delay_140ns();	// Data delay time (read) max 140ns
		data = (uint8_t)LCD_DATA_PIN;
		lcd_en_passive();
	}

	LCD_DIR_WRITE();
	return data;
}


/**
 * Clears the display by setting the whole memory to 0.
 */
void lcd_clear(void) 
{
	uint16_t i;

	if(lcd_mode == LCD_TEXT_MODE) 
	{
		lcd_write_command(0x0A,0x00);
		lcd_write_command(0x0B,0x00);

		for(i = 0; i < LCD_TEXT_LINES * LCD_TEXT_COLUMNS; i++)
			lcd_write_command(0x0C,' ');

		lcd_write_command(0x0A,0x00);
		lcd_write_command(0x0B,0x00);

		lcd_curline = 0;
	}
	else 
	{

		lcd_write_command(0x0A,0x00);
		lcd_write_command(0x0B,0x00);

		for(i = 0; i < ((LCD_GRAPHIC_WIDTH * LCD_GRAPHIC_HEIGHT)/8); i++)
			lcd_write_command(0x0C,0x00);

		lcd_write_command(0x0A,0x00);
		lcd_write_command(0x0B,0x00);
	}


}

void lcd_write_text(char *txt)
{
	uint8_t c,tmp,x,y,ch;


	if(lcd_mode == LCD_TEXT_MODE) 
	{
		c = 0;
		while((ch = *txt) != '\0') 
		{

			if(ch == '\n' || c == LCD_TEXT_COLUMNS) 
			{
				if(lcd_curline < LCD_TEXT_LINES - 1) 
				{
					lcd_curline++;
					c = 0;
					lcd_gotoxy(0,lcd_curline);
				}
				else 
				{
					for(y = 1; y < LCD_TEXT_LINES; y++ ) 
					{
						for(x = 0; x < LCD_TEXT_COLUMNS; x++) 
						{
							lcd_gotoxy(x,y);
							tmp = lcd_read_byte();
							lcd_gotoxy(x,y-1);
							lcd_write_command(0x0C,tmp);
						}
					}
					for(x = 0; x < LCD_TEXT_COLUMNS; x++) 
					{
						lcd_write_command(0x0C,' ');
					}
					lcd_gotoxy(0,LCD_TEXT_LINES-1);
					c = 0;
				}
			}
			else 
			{
				lcd_write_command(0x0C,ch);
				c++;
			}
			txt++;
		}

	}
}



void lcd_write_text_P(const char *txt)
{
	uint8_t c,tmp,x,y,ch;


	if(lcd_mode == LCD_TEXT_MODE) 
	{
		c = 0;
		while((ch = pgm_read_byte(txt)) != '\0') 
		{

			if(ch == '\n' || c == LCD_TEXT_COLUMNS) 
			{
				if(lcd_curline < LCD_TEXT_LINES - 1) 
				{
					lcd_curline++;
					c = 0;
					lcd_gotoxy(0,lcd_curline);
				}
				else 
				{
					for(y = 1; y < LCD_TEXT_LINES; y++ ) 
					{
						for(x = 0; x < LCD_TEXT_COLUMNS; x++) 
						{
							lcd_gotoxy(x,y);
							tmp = lcd_read_byte();
							lcd_gotoxy(x,y-1);
							lcd_write_command(0x0C,tmp);
						}
					}
					for(x = 0; x < LCD_TEXT_COLUMNS; x++) 
					{
						lcd_write_command(0x0C,' ');
					}
					lcd_gotoxy(0,LCD_TEXT_LINES-1);
					c = 0;
				}
			}
			else 
			{
				lcd_write_command(0x0C,ch);
				c++;
			}
			txt++;
		}

	}
}


/**
 * Set the display cursor to a specific location.
 * Right now this does only work in text mode,
 * since there's no need for it now in graphics mode.	\n
 * If the x and y values are to big, the function does nothing.
 *
 * @param x the horizontal cursorposition (starts at 0)
 * @param y the vertical cursorposition (starts at 0)
 *
 */
inline void lcd_gotoxy(uint8_t x, uint8_t y) 
{
	uint16_t pos;

	if(lcd_mode == LCD_TEXT_MODE) 
	{
		if((x < LCD_TEXT_COLUMNS) && (y < LCD_TEXT_LINES)) 
		{
			pos = (y * LCD_TEXT_COLUMNS) + x;

			lcd_write_command(0x0A,(uint8_t) pos );
			lcd_write_command(0x0B,(uint8_t) ((pos & 0xFF00) >> 8));
			lcd_curline = y;
		}
	}
}

/**
 * set or unset a pixel at the specified location
 *
 * @param x horizontal position of the pixel (starts at 0)
 * @param y horizontal position of the pixel (starts at 0)
 * @param set can be PIXEL_ON or PIXEL_OFF
 */
void lcd_plot_pixel(uint8_t x, uint8_t y, uint8_t set) 
{
	if((x < LCD_GRAPHIC_WIDTH) && (y < LCD_GRAPHIC_HEIGHT)) 
	{
		uint16_t pos = y * (LCD_GRAPHIC_WIDTH / 8)  + x / 8;
		lcd_write_command(0x0A,(uint8_t) pos );
		lcd_write_command(0x0B,(uint8_t) ((pos & 0xFF00) >> 8));

		lcd_write_command(set != PIXEL_OFF ? 0x0F : 0x0E, x & 7);
	}
}

/**
 * This function will plot a bitmap with the upper left corner to the given coordinates.\n
 * If the bitmap doesn't fit on the display at this location, nothing is drawn at all.\n
 * Unset pixels are treated as transparent \n
 * This function is by far the most tricky piece of code in this project,\n
 * don't worry if you don't understand it at first sight. \n
 * And if you know a better way to do this, tell me.\n
 *
 * @param x horizontal position of the bitmap
 * @param y vertical position of the bitmap
 * @param bitmap pointer to the bitmap in pgmspace
 * @param w width of the bitmap
 * @param h height of the bitmap
 *
 * This function is dedicated to Greta, one of the most important persons in my life so far.\n
 * Even though I can't remember her actual name.\n
 *
 */
void lcd_plot_bitmap(uint8_t x_off, uint8_t y_off, const uint8_t *bitmap, uint8_t w, uint8_t h) 
{
	uint8_t x, y, cur, curs = 0, sr, p = 0;
	uint16_t pos;

	//check if the bitmap fits on the display
	if((x_off < LCD_GRAPHIC_WIDTH) && (y_off < LCD_GRAPHIC_HEIGHT)
		&& (x_off + w < LCD_GRAPHIC_WIDTH) && (y_off + h < LCD_GRAPHIC_HEIGHT)) 
	{
		//loop linewise through the bitmap
		for(y = y_off; y < y_off + h; y++) 
		{
			cur = 0;
			//loop pixelwise through each line
			for(x = x_off & (uint8_t)0xF8, p = 0; x_off + w > x; x++) 
			{
				if(p == 0)   //load the next byte from display memory
				{
					pos = y * (LCD_GRAPHIC_WIDTH / 8)  + x / 8;
					lcd_write_command(0x0A,(uint8_t) pos );
					lcd_write_command(0x0B,(uint8_t) ((pos & 0xFF00) >> 8));
					cur = lcd_read_byte();
					p = 1;
				}
				if(x >= x_off) 
				{
					sr = (x - x_off) & 7;
					if(sr == 0)   //load the next byte of the bitmap
					{
						curs = pgm_read_byte(bitmap++);
					}
					//grep the pixel from the bitmap and put into the display byte
					if(curs & 1)
						cur |= p; //(1 << dr);
					//else cur &= ~p;				
					curs >>= 1;
				}
				p <<= 1;
				if(p == 0)   //write the current byte to display memory
				{
					pos = y * (LCD_GRAPHIC_WIDTH / 8)  + x / 8;
					lcd_write_command(0x0A,(uint8_t) pos );
					lcd_write_command(0x0B,(uint8_t) ((pos & 0xFF00) >> 8));
					lcd_write_command(0x0C,cur);
				}
			}
			if(p != 0)   //write the last byte to display memory if there's one left
			{
				pos = y * (LCD_GRAPHIC_WIDTH / 8)  + x / 8;
				lcd_write_command(0x0A,(uint8_t) pos );
				lcd_write_command(0x0B,(uint8_t) ((pos & 0xFF00) >> 8));
				lcd_write_command(0x0C,cur);
			}
		}
	}
}

inline void lcd_plot_char(uint8_t x_off, uint8_t y_off, uint8_t c, uint8_t fw, uint8_t fh, const uint8_t* font) 
{
	const uint8_t *letter;
	uint8_t fsize;

	fsize = fh * fw / 8;
	letter = font + c * fsize -1;

	lcd_plot_bitmap(x_off,y_off,letter,fw,fh);
}

void lcd_plot_text(uint8_t x_off, uint8_t y_off, const char *text, uint8_t fw, uint8_t fh, const uint8_t *font) 
{
	while(*text) 
	{
		lcd_plot_char(x_off,y_off,(uint8_t) *text,fw,fh,font);
		x_off += fw;
		text++;
	}
}

void lcd_plot_text_P(uint8_t x_off, uint8_t y_off, const char *text, uint8_t fw, uint8_t fh, const uint8_t *font) 
{
	uint8_t c;

	while((c = pgm_read_byte(text)) != '\0') 
	{
		lcd_plot_char(x_off,y_off,c,fw,fh,font);
		x_off += fw;
		text++;
	}
}


static inline void bresenham_line (uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2)
{	
	int16_t dx, dy, x, y;
	int8_t sdx, sdy;
	uint16_t i, px, py;

	dx = (int16_t)x2-x1;      // the horizontal distance of the line
	dy = (int16_t)y2-y1;      // the vertical distance of the line
	if (dx == 0)
		sdx = 0;
	if (dx > 0)
		sdx = 1;
	else
	{  	sdx = -1;
		dx = -dx;
	}
	if (dy == 0)
		sdy = 0;
	if (dy > 0)
		sdy = 1;
	else
	{	sdy = -1;
		dy = -dy;
	}
	x = dx * 2;
	y = dy * 2;
	px = x1;
	py = y1;
	lcd_setpixel(px, py);

	if (dx >= dy)					 // the line is more horizontal than vertical
	{	
		for(i = 0; i < dx; i++)
		{	
			y += dy;
			if (y >= dx)
			{	
				y -= dx;
				py += sdy;
			}
			px += sdx;
			lcd_setpixel(px, py);
		}
	}
	else				 		// the line is more vertical than horizontal
	{	
		for(i = 0; i < dy; i++)
		{	
			x += dx;
			if (x >= dy)
			{	
				x -= dy;
				px += sdx;
			}
			py += sdy;
			lcd_setpixel(px, py);
		}
	}
}

void lcd_line (uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2)
{   
	if (y1 == y2)
	{	
		if (x1 > x2)
		{	
			uint8_t temp = x1;
			x1 = x2;
			x2 = temp;
		}
		while(x1 <= x2)
			lcd_setpixel(x1++, y1);

	}
	else if (x1 == x2)
	{	
		if (y1 > y2)
		{	
			uint8_t temp = y1;
			y1 = y2;
			y2 = temp;
		}
		while(y1 <= y2)
			lcd_setpixel(x1, y1++);
	}
	else
		bresenham_line(x1, y1, x2, y2);
}

